import time
import graphio
import traceback
from py2neo import Graph
from typing import Dict
from .worker_base import WorkerBase
from .cache_backend import CacheInterface, SetsMetaBase
from linetimer import CodeTimer


class WorkerLoading(WorkerBase):
    def __init__(
        self,
        name,
        set_meta: SetsMetaBase,
        graph_params: Dict,
        insert_action: str = "create",
        create_index: bool = True,
        create_unique_constraints: bool = False,
        amount_of_retries_on_loading_fail: int = 4,
        graphio_batchsize: int = 50000,
    ):
        super(type(self), self).__init__()
        self.name: str = name
        self.set_meta: SetsMetaBase = set_meta
        self.graph_params: Dict = graph_params
        self.insert_action: str = insert_action
        self.create_index = create_index
        self.create_unique_constraints = create_unique_constraints
        self.drain_order_ticket = None
        self.amount_of_retries_on_loading_fail = amount_of_retries_on_loading_fail
        self.graphio_batchsize = graphio_batchsize

    def run(self):
        super(type(self), self).run()
        self.cache: CacheInterface = self.cache_backend(self.cache_backend_params)
        graph_set = None
        log = self.get_logger()
        log.debug(f"START {self.name}")
        try:
            timer = CodeTimer(unit="s", silent=True)
            if self.set_meta.type == graphio.NodeSet:
                with timer:
                    graph_set = self.cache.fetch_NodeSets(self.set_meta)
                log.debug(
                    f"Fetched NodeSet with {len(graph_set.nodes)} Nodes from cache in {timer.took} sec."
                )
            elif self.set_meta.type == graphio.RelationshipSet:
                with timer:
                    graph_set = self.cache.fetch_RelSets(
                        self.set_meta, self.drain_order_ticket
                    )
                log.debug(
                    f"Fetched RelationshipSet with {len(graph_set.relationships)} Relations in {timer.took} sec."
                )
            else:
                log.error(
                    f"Wrong meta data for set. Expected <graphio.NodeSet> or <graphio.RelationshipSet> got {self.set_meta.type}"
                )
                raise ValueError(
                    f"Wrong meta data for set. Expected <graphio.NodeSet> or <graphio.RelationshipSet> got {self.set_meta.type}"
                )
        except Exception as e:
            log.error(traceback.format_exc())
            raise e
        loading_attempt_count = 0
        loading_finished = False
        while not loading_finished:
            try:
                g = Graph(**self.graph_params)
                timer = CodeTimer(unit="s", silent=True)
                if self.create_index:
                    log.debug(f"Create index for {graph_set}")
                    graph_set.create_index(g)
                if self.create_unique_constraints:
                    log.debug(f"Create unique constraint for {graph_set}")
                    raise NotImplementedError
                if self.insert_action == "create":
                    with timer:
                        graph_set.create(g, batch_size=self.graphio_batchsize)
                    log.debug(f"Created {graph_set} into DB in {timer.took} sec")
                elif self.insert_action == "merge":
                    with timer:
                        graph_set.merge(g, batch_size=self.graphio_batchsize)
                    log.debug(f"Merged {graph_set} into DB in {timer.took} sec")
                else:
                    m = f"Unknown insert action. Expected 'create' or 'merge' got '{self.insert_action}'"
                    log.error(m)
                    raise ValueError(m)
                # log loading stats
                if isinstance(graph_set, graphio.NodeSet):
                    amount = f"{len(graph_set.nodes)} nodes"
                    name = f'({":".join(graph_set.labels)})'
                else:
                    amount = f"{len(graph_set.relationships)} relations"
                    name = f"{graph_set.rel_type}"
                if amount == 0:
                    log.warning(f"Worker was tasked with empty graphset {name}")
                log.debug(f"Loaded {name} with {amount} into DB in {timer.took} sec")
                loading_finished = True
            except Exception as e:
                if loading_attempt_count >= self.amount_of_retries_on_loading_fail:
                    log.exception(
                        f"Failed to insert after {loading_attempt_count} tries: {self.set_meta}"
                    )
                    raise e
                loading_attempt_count += 1
                log.warning(
                    f"Failed to {self.insert_action} into neo4j@{self.graph_params} with error \n {str(e)}\n. Will retry {self.amount_of_retries_on_loading_fail - loading_attempt_count + 1} times..."
                )
                time.sleep(20)

        self.cache.report_SetLoaded(graph_set)



import os
import psutil
import graphio
import humanfriendly
from collections import OrderedDict
import time
import graphio
import statistics
import numpy
from typing import TYPE_CHECKING, Type, Union, Dict
from ..worker_base import Progress

if TYPE_CHECKING:
    from .manager import Manager


class Statistics:
    NodeSets_done_timeline: OrderedDict = None
    RelSets_done_timeline: OrderedDict = None
    graph_objects_loaded_counts_per_type_timeline: OrderedDict = None
    graph_objects_cached_counts_per_type_timeline: OrderedDict = None
    graph_objects_count_in_cache: OrderedDict = None
    graph_objects_mean_count_in_cache: OrderedDict = None
    management_ticks_timeline: OrderedDict = None

    def __init__(self, manager):
        self.manager: "Manager" = manager
        self.last_tick_sourcing_workers_done = []
        self.start_time = time.time()

        self.NodeSets_done_timeline = OrderedDict()
        self.RelSets_done_timeline = OrderedDict()
        self.graph_objects_loaded_counts_per_type_timeline = OrderedDict()
        self.graph_objects_cached_counts_per_type_timeline = OrderedDict()
        self.graph_objects_count_in_cache = OrderedDict()
        self.graph_objects_mean_count_in_cache = OrderedDict()
        self.management_ticks_timeline = OrderedDict()

    def update_statistics(
        self, finished_sourcing_workers: list, finished_loading_workers: list
    ):
        current_time = time.time()
        # Save NodeSets/RelSets Workers timeline
        self.NodeSets_done_timeline[current_time] = [
            w for w in finished_loading_workers if w.set_meta.type == graphio.NodeSet
        ]

        self.RelSets_done_timeline[current_time] = [
            w
            for w in finished_loading_workers
            if w.set_meta.type == graphio.RelationshipSet
        ]

        # Save Labels/Rels amount loaded
        self.graph_objects_loaded_counts_per_type_timeline[
            current_time
        ] = self.manager.cache.get_report_graphsets_loaded()

        # Save Labels/Rels amount loaded
        self.graph_objects_cached_counts_per_type_timeline[
            current_time
        ] = self.manager.cache.get_report_graphsets_cached()

        self.graph_objects_count_in_cache[current_time] = (
            self.get_cached_graphobjects_total_count()
            - self.get_loaded_graphobjects_total_count()
        )

        self.graph_objects_mean_count_in_cache[
            current_time
        ] = self.get_mean_count_graph_objects_in_cache()

        # Save amount of management ticks
        self.management_ticks_timeline[current_time] = self.manager.management_ticks

    def get_management_ticks_per_n_sec(
        self, n: int = 1, timeframe: int = 120, object_type=None
    ) -> int:
        current_time = time.time()
        running_time = current_time - self.start_time

        if timeframe is None or running_time < timeframe:
            timeframe_start = self.start_time
        else:
            timeframe_start = current_time - timeframe

        relevant_data = OrderedDict()
        for timestamp, tick_count in self.management_ticks_timeline.items():
            if timestamp > timeframe_start:
                relevant_data[timestamp] = tick_count
        relevnt_tick_counts = list(relevant_data.values())
        if len(relevnt_tick_counts) < 2:
            tick_count = 0
        else:
            tick_count = (
                relevnt_tick_counts[len(relevant_data) - 1] - relevnt_tick_counts[0]
            )
        return (tick_count / (current_time - timeframe_start)) * n

    def get_graphobjects_count_loaded_per_n_sec(
        self, n: int = 1, timeframe: int = 120, object_type: str = None
    ) -> int:
        """[summary]

        Args:
            n (int, optional): Timeresolution in sec. Defaults to 1.
            timeframe (int, optional): Take last n seconds into account to calculate median value. Defaults to 20.
            object_type (str, optional): "Nodes" to only count Nodes per second, "Relations" to only count rels per second or None to count both. Defaults to None.

        Returns:
            int: Integer as objects per n seconds loaded
        """
        return self._get_graphobjects_count_per_n_sec(
            n=n, timeframe=timeframe, object_type=object_type, target="Loaded"
        )

    def get_graphobjects_count_cached_per_n_sec(
        self, n: int = 1, timeframe: int = 120, object_type: str = None
    ) -> int:
        """[summary]

        Args:
            n (int, optional): Timeresolution in sec. Defaults to 1.
            timeframe (int, optional): Take last n seconds into account to calculate median value. Defaults to 20.
            object_type (str, optional): "Nodes" to only count Nodes per second, "Relations" to only count rels per second or None to count both. Defaults to None.

        Returns:
            int: Integer as objects per n seconds loaded
        """
        return self._get_graphobjects_count_per_n_sec(
            n=n, timeframe=timeframe, object_type=object_type, target="Cached"
        )

    def _get_graphobjects_count_per_n_sec(
        self,
        n: int = 1,
        timeframe: int = 120,
        object_type: str = None,
        target: str = "Loaded",
    ) -> int:
        """[summary]

        Args:
            n (int, optional): Timeresolution in sec. Defaults to 1.
            timeframe (int, optional): Take last n seconds into account to calculate median value. Defaults to 20.
            object_type (str, optional): "Nodes" to only count Nodes per second, "Relations" to only count rels per second or None to count both. Defaults to None.
            target (str, optional): Do you want to count graphsets loaded into the neo4j DB use "Loaded" or do you want to count cached graphobjects into the cache backend (e.g. redis) use "Cached". Defaults to "Loaded"
        Returns:
            int: Integer as objects per n seconds loaded
        """
        current_time = time.time()
        running_time = current_time - self.start_time
        if timeframe is None or running_time < timeframe:
            timeframe_start = self.start_time
        else:
            timeframe_start = current_time - timeframe
        relevant_data = OrderedDict()

        for timestamp, data in (
            self.graph_objects_loaded_counts_per_type_timeline.items()
            if target == "Loaded"
            else self.graph_objects_cached_counts_per_type_timeline.items()
        ):
            if timestamp > timeframe_start:
                relevant_data[timestamp] = data
        count_before_and_after = [0, 0]
        if len(relevant_data) == 0:
            return 0
        for index, data_position in enumerate([0, len(relevant_data) - 1]):
            count_all = 0
            report = list(relevant_data.values())[data_position]
            if object_type in [None, "Nodes"]:
                for graphobject_name, count in report["Nodes"].items():
                    count_all += count
            if object_type in [None, "Relations"]:
                for graphobject_name, count in report["Relations"].items():
                    count_all += count
            count_before_and_after[index] = count_all
        if len(relevant_data) < 2:
            count_before_and_after[0] = 0
        return ((count_before_and_after[1] - count_before_and_after[0]) / timeframe) * n

    def get_count_graphsets_loaded_per_n_sec(
        self,
        n: int = 1,
        timeframe: int = 120,
        graphset_type: Type[Union[graphio.NodeSet, graphio.RelationshipSet]] = None,
    ):
        current_time = time.time()
        running_time = current_time - self.start_time
        if timeframe is None or running_time < timeframe:
            timeframe_start = self.start_time
        else:
            timeframe_start = current_time - timeframe

        workers = []
        for timestamp, workers_in_slot in self.NodeSets_done_timeline.items():
            if timestamp > timeframe_start:
                if graphset_type is None or graphset_type == graphio.NodeSet:
                    workers.extend(workers_in_slot)
                if graphset_type is None or graphset_type == graphio.RelationshipSet:
                    # this is a little bit hacky but we know for sure that the keys in self.RelSets_done_timeline and self.NodeSets_done_timeline are always in sync
                    workers.extend(self.RelSets_done_timeline[timestamp])
        return (len(workers) / (current_time - timeframe_start)) * n

    def get_mean_count_graph_objects_in_cache(
        self,
        timeframe: int = 120,
        graphset_type: Type[Union[graphio.NodeSet, graphio.RelationshipSet]] = None,
    ) -> int:
        current_time = time.time()
        running_time = current_time - self.start_time

        if timeframe is None or running_time < timeframe:
            timeframe_start = self.start_time
        else:
            timeframe_start = current_time - timeframe

        relevant_data = []
        for (
            timestamp,
            graph_objects_count,
        ) in self.graph_objects_count_in_cache.items():
            if timestamp > timeframe_start:
                relevant_data.append(graph_objects_count)
        return round(statistics.mean(relevant_data))

    def get_count_running_sourcing_workers(self) -> int:
        return len(
            self.manager.manager_sourcing._get_workers(status=("initial", "running"))
        )

    def get_count_left_sourcing_workers(self) -> int:
        return len(self.manager.manager_sourcing.workers) - len(
            self.manager.manager_sourcing._get_workers(progress="complete")
        )

    def get_memory_consumed_by_loaders(self):
        mem_total_bytes = 0
        for worker in self.manager.manager_loading._get_workers(status="running"):
            process = psutil.Process(worker.pid)
            mem_total_bytes += process.memory_info()[0]
        return mem_total_bytes

    def get_memory_consumed_by_sourcing_workers(self):
        mem_total_bytes = 0
        for worker in self.manager.manager_sourcing._get_workers(status="running"):
            process = psutil.Process(worker.pid)
            mem_total_bytes += process.memory_info()[0]
        return mem_total_bytes

    def get_memory_consumed_by_manager(self):
        process = psutil.Process(os.getpid())
        return process.memory_info()[0]

    def is_sourcing_phase_done(self):
        return self.manager.manager_sourcing.is_done()

    def get_memory_available(self):
        return getattr(psutil.virtual_memory(), "available")

    def get_memory_total(self):
        return getattr(psutil.virtual_memory(), "total")

    def get_memory_used(self):
        return getattr(psutil.virtual_memory(), "used")

    def get_cache_storage_total(self) -> int:
        return self.manager.strategy.cache_storage_total

    def get_cache_storage_available(self) -> int:
        return self.get_cache_storage_total() - self.get_cache_storage_used()

    def get_cache_storage_used(self) -> int:
        return self.manager.cache.storage_used()

    def get_cached_graphobjects_total_count(self, object_type: str = None) -> int:
        if not self.graph_objects_cached_counts_per_type_timeline:
            return 0
        last_count_stats = self.graph_objects_cached_counts_per_type_timeline[
            next(reversed(self.graph_objects_cached_counts_per_type_timeline))
        ]
        total_count = 0
        if object_type in [None, "Relations"]:
            for label, count in last_count_stats["Relations"].items():
                total_count += count
        if object_type in [None, "Nodes"]:
            for label, count in last_count_stats["Nodes"].items():
                total_count += count
        return total_count

    def get_loaded_graphobjects_total_count(self, object_type: str = None) -> int:
        if not self.graph_objects_loaded_counts_per_type_timeline:
            return 0
        last_count_stats = self.graph_objects_loaded_counts_per_type_timeline[
            next(reversed(self.graph_objects_loaded_counts_per_type_timeline))
        ]
        total_count = 0
        if object_type in [None, "Relations"]:
            for label, count in last_count_stats["Relations"].items():
                total_count += count
        if object_type in [None, "Nodes"]:
            for label, count in last_count_stats["Nodes"].items():
                total_count += count
        return total_count

    def get_cached_graphobjects_total_count_trend(
        self,
        timeframe: int = 120,
    ) -> int:
        current_time = time.time()
        running_time = current_time - self.start_time

        if timeframe is None or running_time < timeframe:
            timeframe_start = self.start_time
        else:
            timeframe_start = current_time - timeframe

        relevant_data = []
        for (
            timestamp,
            graph_objects_count,
        ) in self.graph_objects_mean_count_in_cache.items():
            if timestamp > timeframe_start:
                relevant_data.append(graph_objects_count)
        return self._trend(relevant_data)

    def get_cached_graphobjects_current(self) -> int:
        raise NotImplementedError

    def get_cache_storage_level(self) -> str:
        mem_full_perc = (
            self.get_cache_storage_used() * 100
        ) / self.get_cache_storage_total()
        if self.manager.strategy.cache_storage_clogged_limit <= mem_full_perc:
            return "RED"
        elif self.manager.strategy.cache_storage_warning_limit <= mem_full_perc:
            return "ORANGE"
        else:
            return "GREEN"

    def get_cache_storage_used_by_relSets(self):
        size = 0
        for rel_meta in self.manager.cache.list_SetsMeta(
            set_type=graphio.RelationshipSet
        ):
            size += rel_meta.total_size_bytes
        return size

    def get_cache_storage_used_by_nodeSets(self):
        size = 0
        for rel_meta in self.manager.cache.list_SetsMeta(set_type=graphio.NodeSet):
            size += rel_meta.total_size_bytes
        return size

    def get_report(self):
        report = {}

        report["General"] = {}

        gen_rep = report["General"]

        gen_rep["running_time_sec"] = {
            "val": time.time() - self.start_time,
            "desc": "Running time",
            "humanfriendly_transformer": humanfriendly.format_timespan,
        }

        gen_rep["max_processes_count"] = {
            "val": self.manager.strategy.max_processes_count,
            "desc": "Max processes count",
        }

        gen_rep["management_tick_per_sec"] = {
            "val": self.get_management_ticks_per_n_sec(),
            "desc": "Management ticks per second",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        report["Sourcing"] = {}

        src_rep = report["Sourcing"]

        src_rep["cores_avail_no"] = {
            "val": self.manager.strategy.amount_sourcing_cores(),
            "desc": "Sourcing Workers amount of cores allocated",
        }
        src_rep["workers_total_no"] = {
            "val": self.manager.manager_sourcing.worker_count_total,
            "desc": "Sourcing Workers amount total",
        }

        src_rep["workers_fin_no"] = {
            "val": len(self.manager.manager_sourcing.finished_workers),
            "desc": "Sourcing Workers amount finished",
        }

        src_rep["workers_running_no"] = {
            "val": len(self.manager.manager_sourcing._get_workers(status="running")),
            "desc": "Sourcing Workers amount running",
        }

        report["Cache"] = {}

        cache_rep = report["Cache"]

        cache_rep["relSet_no"] = {
            "val": len(self.manager.cache.list_SetsMeta(graphio.RelationshipSet)),
            "desc": "RelationSets count in cache",
        }
        cache_rep["relSet_size"] = {
            "val": self.get_cache_storage_used_by_relSets(),
            "desc": "RelationSets size in cache",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        cache_rep["rels_per_sec"] = {
            "val": self.get_graphobjects_count_cached_per_n_sec(
                object_type="Relations"
            ),
            "desc": "Relations cached per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        cache_rep["nodeSet_no"] = {
            "val": len(self.manager.cache.list_SetsMeta(graphio.NodeSet)),
            "desc": "NodeSets count in cache",
        }
        cache_rep["nodeSet_size"] = {
            "val": self.get_cache_storage_used_by_nodeSets(),
            "desc": "NodeSets size in cache",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        cache_rep["nodes_per_sec"] = {
            "val": self.get_graphobjects_count_cached_per_n_sec(object_type="Nodes"),
            "desc": "Nodes cached per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }
        #

        cache_rep["cache_storage_status"] = {
            "val": self.get_cache_storage_level(),
            "desc": "Cache storage health status",
        }

        cache_storage_level_details = f"{int(100 - ((self.get_cache_storage_used() * 100) / self.get_cache_storage_total()))}% Free ({humanfriendly.format_size(self.get_cache_storage_used())} used of {humanfriendly.format_size(self.get_cache_storage_total())})"

        cache_rep["cache_storage_status_details"] = {
            "val": cache_storage_level_details,
            "desc": "Cache storage status details",
        }

        report["Loading"] = {}

        load_rep = report["Loading"]

        load_rep["cores_avail_no"] = {
            "val": self.manager.strategy.amount_loading_cores(),
            "desc": "Loading Workers amount of cores allocated",
        }
        load_rep["workers_running_no"] = {
            "val": len(self.manager.manager_loading._get_workers(status="running")),
            "desc": "Loading Workers amount running",
        }
        load_rep["workers_fin_no"] = {
            "val": len(self.manager.manager_loading.finished_workers),
            "desc": "Loading Workers amount finished",
        }

        load_rep["workers_relsets_run_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    status=("running"),
                    tag=self.manager.manager_loading.worker_tag_relsets,
                )
            ),
            "desc": "Loading RelSets Workers amount running",
        }

        load_rep["workers_relsets_fin_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    progress=Progress.COMPLETE,
                    tag=self.manager.manager_loading.worker_tag_relsets,
                )
            ),
            "desc": "Loading RelSets Workers amount finished",
        }

        load_rep["workers_relsets_wait_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    progress=Progress.DRAIN_ORDERED,
                    tag=self.manager.manager_loading.worker_tag_relsets,
                )
            ),
            "desc": "RelSets Workers waiting for NodeSet drain",
        }

        load_rep["labels_blocked_for_rels"] = {
            "val": len(self.manager._get_blocked_labels()),
            "desc": "Amount of labels blocked for relation loading",
        }

        load_rep["relSet_per_sec"] = {
            "val": self.get_count_graphsets_loaded_per_n_sec(
                graphset_type=graphio.RelationshipSet
            ),
            "desc": "RelationSets loaded per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        load_rep["rels_per_sec"] = {
            "val": self.get_graphobjects_count_loaded_per_n_sec(
                object_type="Relations"
            ),
            "desc": "Relations loaded per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        load_rep["workers_relsets_run_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    status=("running"),
                    tag=self.manager.manager_loading.worker_tag_relsets,
                )
            ),
            "desc": "Loading RelSets Workers amount running",
        }

        load_rep["workers_NodeSets_run_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    progress=(Progress.QUEUED, Progress.LOADING),
                    tag=self.manager.manager_loading.worker_tag_nodesets,
                )
            ),
            "desc": "Loading NodeSets Workers running or queded",
        }

        load_rep["workers_NodeSets_fin_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    progress=Progress.COMPLETE,
                    tag=self.manager.manager_loading.worker_tag_nodesets,
                )
            ),
            "desc": "Loading NodeSets Workers amount finished",
        }

        load_rep["nodeSet_per_sec"] = {
            "val": self.get_count_graphsets_loaded_per_n_sec(
                graphset_type=graphio.NodeSet
            ),
            "desc": "NodeSets loaded per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        load_rep["nodes_per_sec"] = {
            "val": self.get_graphobjects_count_loaded_per_n_sec(object_type="Nodes"),
            "desc": "Nodes loaded per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        report["SyncControl"] = {}
        sync_rep = report["SyncControl"]

        sync_rep["sourcing_paused"] = {
            "val": str(not self.manager.cache.store_is_available()),
            "desc": "Sourcing is paused to let loading catch up",
        }
        sync_rep["cached_total"] = {
            "val": self.get_cached_graphobjects_total_count(),
            "desc": "Count of Graphobjects inserted into cache",
            "humanfriendly_transformer": humanfriendly.format_number,
        }

        sync_rep["loaded_total"] = {
            "val": self.get_loaded_graphobjects_total_count(),
            "desc": "Count of Graphobjects loaded into Neo4j",
            "humanfriendly_transformer": humanfriendly.format_number,
        }

        sync_rep["cache_to_load_advantage_count_current"] = {
            "val": sync_rep["cached_total"]["val"] - sync_rep["loaded_total"]["val"],
            "desc": "Current exceeding cached graphobjects compared to loaded",
            "humanfriendly_transformer": humanfriendly.format_number,
        }

        sync_rep["cached_graphobjects_count"] = {
            "val": self.get_mean_count_graph_objects_in_cache(),
            "desc": "Mean cached graphobjects count",
            "humanfriendly_transformer": humanfriendly.format_number,
        }

        sync_rep["max_cached_graphobjects_count"] = {
            "val": self.manager.strategy.max_graphobjects_count_in_cache,
            "desc": "Max graphobjects cache count before pausing sourcing workers",
            "humanfriendly_transformer": humanfriendly.format_number,
        }

        sync_rep["min_cached_graphobjects_count"] = {
            "val": self.manager.strategy.min_graphobjects_count_in_cache,
            "desc": "Min graphobjects cache count",
            "humanfriendly_transformer": humanfriendly.format_number,
        }
        sync_rep["cache_to_load_advantage"] = {
            "val": (
                (
                    (sync_rep["cached_total"]["val"] / sync_rep["loaded_total"]["val"])
                    * 100
                )
                - 100
            )
            if sync_rep["loaded_total"]["val"] not in [None, 0]
            else 0,
            "desc": "Percental advantage of cached to loaded graphobjects",
            "humanfriendly_transformer": "{:.2f} %".format,
        }

        sync_rep["cache_level_trend"] = {
            "val": self.get_cached_graphobjects_total_count_trend(),
            "desc": "Trend of cache level (+=filling,-=drying)",
        }

        report["Memory"] = {}
        mem_rep = report["Memory"]

        mem_rep = report["Memory"]

        memory_status = f"{humanfriendly.format_size(self.get_memory_available())} available - {humanfriendly.format_size(self.get_memory_used())} used of {humanfriendly.format_size(self.get_memory_total())}"

        mem_rep["memory_status_details"] = {
            "val": memory_status,
            "desc": "Memory status for Manager and Workers processes",
        }

        mem_rep["memory_used_by_manager"] = {
            "val": self.get_memory_consumed_by_manager(),
            "desc": "Memory used by the manager process",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        mem_rep["memory_used_by_sourcing_workers"] = {
            "val": self.get_memory_consumed_by_sourcing_workers(),
            "desc": "Memory used by the sourcing processes",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        mem_rep["memory_used_by_loading_workers"] = {
            "val": self.get_memory_consumed_by_loaders(),
            "desc": "Memory used by the loading processes",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        return report

    def get_human_readable_report(self):

        report = self.get_report()

        report_str = "\n"
        for title, chapters in report.items():
            report_str += f"## {title}\n"
            longest_chapter_desc = max(
                [len(chapter["desc"]) for chapter in chapters.values()]
            )
            for key, chapter in chapters.items():
                if "humanfriendly_transformer" in chapter:
                    val = chapter["humanfriendly_transformer"](chapter["val"])
                else:
                    val = chapter["val"]
                row = "\t{:" + str(longest_chapter_desc) + "}: {}\n"
                report_str += row.format(chapter["desc"], val)

        return report_str

    @classmethod
    def _trend(cls, data: list) -> int:
        if len(data) > 1:
            coeffs = numpy.polynomial.polynomial.Polynomial.fit(
                range(0, len(data)), list(data), 1
            )
            coeffs_convert = coeffs.convert()
            slope = list(coeffs_convert)[-1]
            return round(float(slope), 2)
        else:
            return 0



import logging
import graphio
from typing import List, Dict, Type, Union, Tuple, Set, TYPE_CHECKING
from ..worker_base import WorkerBase, Progress
from .manager_workers_base import ManagerWorkersBase
from ..cache_backend import (
    NodeSetMeta,
    RelationSetMeta,
    SetsMetaBase,
)

if TYPE_CHECKING:
    from .manager import Manager

log = logging.getLogger(__name__)


class ManagerWorkersLoading(ManagerWorkersBase):
    def __init__(self, manager: "Manager", worker_class: Type[WorkerBase]):
        super(type(self), self).__init__(manager, worker_class)
        self.parent: "Manager" = manager

        # Setup caching backend on worker class
        self.parent.worker_loading_class.cache_backend = self.parent.cache_backend
        self.parent.worker_loading_class.cache_backend_params = (
            self.parent.cache_backend_params
        )
        self.worker_tag_relsets = "RELSET"
        self.worker_tag_nodesets = "NODESET"

    def manage(self):
        c = self.parent.strategy.amount_loading_nodes_cores()
        self.manage_nodeset_loading(c)
        c = self.parent.strategy.amount_loading_rels_cores()
        self.manage_relation_loading(c)

    def is_done(self):
        # we are done when:
        # sourcing is finished and there is no more data in the cache...
        if (
            self.parent.manager_sourcing.is_done()
            and not self.parent.cache.list_SetsMeta()
        ):
            # ... and all workers are finished
            if (
                len(self._get_workers(status="running"))
                + len(self._get_workers(status="initial"))
                == 0
            ):
                return True
        else:
            return False

    def manage_nodeset_loading(self, assigned_no_of_cores):
        log.debug("###MANAGE NODESETS###")
        cached_sets = self.parent.cache.list_SetsMeta(set_type=graphio.NodeSet)
        # find NodeSets that have workers assigned (allready running or queued)
        cached_sets_worked_on: List[SetsMetaBase] = [
            worker.set_meta
            for worker in self._get_workers(
                progress=(Progress.QUEUED, Progress.LOADING),
                tag=self.worker_tag_nodesets,
            )
        ]
        # find nodeSets types that are not loaded atm
        cached_sets_not_worked_on: List[SetsMetaBase] = [
            cs for cs in cached_sets if cs not in cached_sets_worked_on
        ]
        # Collect nodeSets that are not allready loading and that are not blocked. These are the NodeSet types we can tackle next
        cached_sets_not_worked_on_and_not_blocked: List[SetsMetaBase] = [
            cs
            for cs in cached_sets_not_worked_on
            if not self.parent._is_nodeset_loading_blocked(cs)
        ]
        if cached_sets_not_worked_on_and_not_blocked:

            no_of_free_cores = assigned_no_of_cores - len(
                self._get_workers(status="running", tag=self.worker_tag_nodesets)
            )

            if no_of_free_cores < 0:
                no_of_free_cores = 0
            chached_sets_next = cached_sets_not_worked_on_and_not_blocked[
                :no_of_free_cores
            ]

            # Prepare parameter to initialize workers
            workers_params: List[Dict] = [
                {
                    "set_meta": cached_set_meta,
                    "graph_params": self.parent.graph_params,
                    "insert_action": self.parent.insert_action,
                    "create_index": self.parent.create_indexes,
                    "create_unique_constraints": self.parent.create_unique_constraints,
                    "graphio_batchsize": self.parent.strategy.graphio_batchsize,
                }
                for cached_set_meta in chached_sets_next
            ]
            # generate human readable names for workers
            worker_names: List[str] = [
                ":".join(self.parent.cache.get_NodeSetMeta(cached_set_meta).labels)
                for cached_set_meta in chached_sets_next
            ]
            # Initialize new workers
            self._init_workers(
                parameters=workers_params,
                names=worker_names,
                tags=[self.worker_tag_nodesets],
            )
        no_of_free_cores = assigned_no_of_cores - len(
            self._get_workers(status="running", tag=self.worker_tag_nodesets)
        )

        #
        # log.info(f"no_of_free_cores: {no_of_free_cores}/{assigned_no_of_cores}")

        # start new workers:
        for next_worker in self._get_workers(
            progress=Progress.QUEUED, tag=self.worker_tag_nodesets
        ):
            next_worker.progress = Progress.LOADING
            next_worker.timer.__enter__()
            next_worker.start()

        # collect exited workers
        workers_did_finished = self._finish_workers(self.worker_tag_nodesets)

    def manage_relation_loading(self, assigned_no_of_cores):
        log.debug(
            f"Manage Relsets loading with {assigned_no_of_cores} cores available"
            + f" and {len(self._get_workers(status='running',tag=self.worker_tag_relsets))} worker running atm."
            + f" {len(self._get_workers(progress=Progress.DRAIN_ORDERED, tag=self.worker_tag_relsets))} workers are waiting for a drain"
        )

        cached_sets = self.parent.cache.list_SetsMeta(set_type=graphio.RelationshipSet)

        # find RelSets that have workers assigned (allready running or queued)
        cached_sets_worked_on: List[RelationSetMeta] = [
            self.parent.cache.get_RelSetMeta(worker.set_meta)
            for worker in self._get_workers(
                progress=(
                    Progress.QUEUED,
                    Progress.DRAIN_ORDERED,
                    Progress.DRAIN_READY,
                    Progress.LOADING,
                ),
                tag=self.worker_tag_relsets,
            )
        ]

        log.debug(
            f"Relations cached_sets_worked_on count: {len(cached_sets_worked_on)}"
        )
        if len(cached_sets_worked_on) < assigned_no_of_cores:
            cached_sets_not_worked_on: List[RelationSetMeta] = [
                self.parent.cache.get_RelSetMeta(cs)
                for cs in cached_sets
                if cs not in cached_sets_worked_on
            ]

            if cached_sets_not_worked_on:
                no_of_free_cores = assigned_no_of_cores - len(cached_sets_worked_on)
                cached_sets_next = []
                for next_set in cached_sets_not_worked_on:
                    # filter out overlapping nodesets, which could cause nodelocks
                    overlaps = False
                    # get target labels of the potential next relationship loader
                    next_ns_labels = (
                        next_set.end_node_labels + next_set.start_node_labels
                    )
                    # get all target labels that currently affected by running relationship loaders
                    current_ns_labels = {
                        label
                        for cs in cached_sets_worked_on + cached_sets_next
                        for label in cs.start_node_labels + cs.end_node_labels
                    }

                    for n_label in next_ns_labels:
                        if n_label in current_ns_labels:
                            overlaps = True

                    if not overlaps:
                        cached_sets_next.append(next_set)
                    if len(cached_sets_next) >= no_of_free_cores:
                        break

                # cached_sets_next = cached_sets_not_worked_on[:no_of_free_cores]
                # Prepare parameter and name list to initialize new workers
                workers_params: List[Dict] = [
                    {
                        "set_meta": cached_set_meta,
                        "graph_params": self.parent.graph_params,
                        "insert_action": self.parent.insert_action,
                        "create_index": self.parent.create_indexes,
                        "create_unique_constraints": self.parent.create_unique_constraints,
                    }
                    for cached_set_meta in cached_sets_next
                ]
                # generate human readable names for workers
                worker_names: List[str] = []
                for cached_set_meta in cached_sets_next:
                    worker_names.append(cached_set_meta.rel_type)
                # Initialize workers
                self._init_workers(
                    parameters=workers_params,
                    names=worker_names,
                    tags=[self.worker_tag_relsets],
                )

        # manage drain orders
        self._order_drains(assigned_no_of_cores)
        self._check_for_drains_ready(assigned_no_of_cores)

        # start one new worker per management tick
        for next_worker in self._get_workers(
            progress=Progress.DRAIN_READY, tag=self.worker_tag_relsets
        ):
            next_worker.progress = Progress.LOADING
            next_worker.timer.__enter__()
            log.debug(f"Start {next_worker.name}")
            next_worker.start()
            # we break the loop here as we only start one relationshipSet loader per management tick
            # break

        # collect exited workers
        self._finish_workers(self.worker_tag_relsets)

    def _order_drains(self, assigned_no_of_cores):

        # Order drain for relationsSet attached NodeSets for next waiting worker

        next_relset_workers = [
            w
            for w in self._get_workers(
                progress=Progress.QUEUED, tag=self.worker_tag_relsets
            )
        ]
        for next_worker in next_relset_workers[:1]:
            drain_order_ticket_id = self.parent.cache.order_RelSetDrain(
                relset_meta=next_worker.set_meta
            )
            next_worker.progress = Progress.DRAIN_ORDERED
            next_worker.drain_order_ticket = drain_order_ticket_id

    def _check_for_drains_ready(self, assigned_no_of_cores):
        # Check if any running drains are ready
        for drain_wait_worker in self._get_workers(
            progress=Progress.DRAIN_ORDERED, tag=self.worker_tag_relsets
        ):
            if not self.parent.cache.is_drain_done(
                drain_wait_worker.drain_order_ticket
            ):
                # The drain is still running in the cache. we cant do anything atm
                # lets skip to the next relationSet
                continue

            # drain is ready: now we need to block new workers on our target nodesets labels and wait for running workers, that are loading our target nodeset types, to finish
            self.parent._block_loading_label(
                drain_wait_worker.id,
                set(
                    drain_wait_worker.set_meta.start_node_labels
                    + drain_wait_worker.set_meta.end_node_labels
                ),
            )

            # It can happen that workers, which were started before the drain order and are loading nodesets that are related to the relationshipSet, are still loading.
            # this means: there could be nodelocks when loading the relationshipSet or worse target nodes could be missing.
            # We need to make sure that there are no old running nodeset workers affecting/occupying our relationsship target nodesets
            nodesets_currently_worked_on: List[NodeSetMeta] = [
                self.parent.cache.get_NodeSetMeta(worker.set_meta)
                for worker in self._get_workers(
                    progress=(Progress.QUEUED, Progress.LOADING),
                    tag=self.worker_tag_nodesets,
                )
            ]
            labels_of_nodesets_currently_worked_on: Set[str] = {
                label for cs in nodesets_currently_worked_on for label in cs.labels
            }

            nodesets_occupied = False
            for label in (
                drain_wait_worker.set_meta.start_node_labels
                + drain_wait_worker.set_meta.end_node_labels
            ):
                if label in labels_of_nodesets_currently_worked_on:
                    # The drain ready but there are still other workers randomly working on this nodeset. we cant do anything atm
                    nodesets_occupied = True
            if nodesets_occupied:
                # lets skip to the next relationSet
                continue
            log.debug(
                f"DRAIN SEEMS READY drain_wait_worker.drain_order_ticket: {drain_wait_worker.drain_order_ticket}"
            )
            # Drain is ready and nothing is in the way of loading the relationshipset to the DB now. lets mark the worker as ready
            drain_wait_worker.progress = Progress.DRAIN_READY



import operator
import os
import psutil
import graphio
import math
import humanfriendly
import time
import multiprocessing
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .manager import Manager


class StrategyBase:
    max_processes_count: int = multiprocessing.cpu_count()
    cache_storage_warning_limit: int = 80
    cache_storage_clogged_limit: int = 90
    max_graphobjects_count_in_cache: int = 1000000
    min_graphobjects_count_in_cache: int = None
    target_graphobjects_count_in_cache: int = None
    graphio_batchsize: int = 50000

    cache_storage_total = None

    # On large server with huge core counts, the sourcing workers can fly the cache quickly into the ceilling. thats why we create a default max amount of sourcing workers
    max_sourcing_workers = None

    def __init__(self, manager: "Manager"):
        self.manager = manager
        self._cache_untouched = True
        self._cache_first_hit_time = 0
        if self.cache_storage_total is None:
            self.cache_storage_total = getattr(psutil.virtual_memory(), "total") * 0.80
        else:
            self.cache_storage_total = humanfriendly.parse_size(
                self.cache_storage_total
            )
        if not self.target_graphobjects_count_in_cache:
            self.target_graphobjects_count_in_cache = round(
                self.max_graphobjects_count_in_cache * 0.5
            )
        if not self.min_graphobjects_count_in_cache:
            self.min_graphobjects_count_in_cache = round(
                self.max_graphobjects_count_in_cache * 0.1
            )

    def amount_loading_cores(self) -> int:
        if (
            self.manager.statistics.is_sourcing_phase_done()
            or self.manager.statistics.get_cache_storage_used()
            >= self.manager.cache_size
            or self.manager.statistics.get_cache_storage_level() in ["ORANGE", "RED"]
        ):
            return self.max_processes_count
        c = round(self.max_processes_count - self.amount_sourcing_cores())
        if c == 0:
            c = 1
        # We can take more cores as available because loading processes are waiting a lot for the db and wont use any cpu time during that
        return math.ceil(c * 1.5)

    def amount_loading_nodes_cores(self) -> int:
        c = round(self.amount_loading_cores() * 0.9)
        if c == 0:
            c = 1
        # We can take more cores as available because loading processes are waiting a lot for the db and wont use any cpu time during that
        return math.ceil(c * 1.5)

    def amount_loading_rels_cores(self) -> int:
        # if we ended the sourcing phase (all sourcing workers finished) we first load all nodesets and after that all relsets can load without draining)
        if (
            self.manager.statistics.get_count_left_sourcing_workers() == 0
            and len(self.manager.cache.list_SetsMeta(graphio.NodeSet)) > 0
        ):
            return 0
        # if we are low on nodeSets in the cache we can start more RelSet loaders
        nodesets_cores_needed: int = len(
            self.manager.cache.list_SetsMeta(graphio.NodeSet)
            + self.manager.manager_loading._get_workers(
                status=("initial", "running"),
                tag=self.manager.manager_loading.worker_tag_nodesets,
            )
        )
        if nodesets_cores_needed < self.amount_loading_nodes_cores():
            return self.amount_loading_cores() - nodesets_cores_needed
            # else
        # By default we apply 50% of available loading cores to RelSets (or at least one Core)
        c = round(self.amount_loading_cores() * 0.5)
        if c == 0:
            c = 1
        return math.ceil(c)

    def amount_sourcing_cores(self) -> int:
        c = 0
        if self._cache_untouched:
            # when nothing is in cache yet we only let one parser work.
            # we want to measure how long the first parser need to load data into the cache
            # this way we can carefully increase parser workers without clogging the cache
            if self.manager.statistics.get_cached_graphobjects_total_count() > 0:
                # first objects arrive in cache.
                # from here on we can increase the amount of parsing workers
                self._cache_untouched = False
                self._cache_first_hit_time = (
                    self.manager.statistics.start_time - time.time()
                )
                if self._cache_first_hit_time < 5:
                    # we want at least a cooldown time of 5 sec for spawning new parsers
                    # otherwise we take the time the first parser needed to fill first data into the cache
                    self._cache_first_hit_time = 3
                self._sourcing_cool_down_target_time = (
                    time.time() + self._cache_first_hit_time
                )
            c = 1
        else:
            if self.manager.statistics.get_cache_storage_level() == "RED":
                # if memory is full, dont start any new parsers. we wait until the situation is better
                return 0
            if self._sourcing_cool_down_target_time < time.time():
                # Reset timer
                self._sourcing_cool_down_target_time = (
                    time.time() + self._cache_first_hit_time
                )
                cache_level_mean = (
                    self.manager.statistics.get_mean_count_graph_objects_in_cache()
                )
                cache_level_current = (
                    self.manager.statistics.get_cached_graphobjects_total_count()
                    - self.manager.statistics.get_loaded_graphobjects_total_count()
                )

                cache_level_trend = (
                    self.manager.statistics.get_cached_graphobjects_total_count_trend()
                )
                # lets have a look at the trend of the cache. it is filling or drying? Based on that we give more or less cores for parsers
                if cache_level_trend < 100:
                    c = self._last_amount_sourcing_cores + 1
                elif cache_level_trend > 100:
                    c = self._last_amount_sourcing_cores - 1
                else:
                    # no trend identifiable.
                    if (
                        cache_level_current < self.min_graphobjects_count_in_cache
                        or cache_level_mean < self.min_graphobjects_count_in_cache
                    ):
                        # oh we scrape the bottom of the cache. That could distort our trend analyses and is not where we want to be anyway.
                        # turn up the volume
                        c = self._last_amount_sourcing_cores + 1
                    else:
                        # lets do nothing and wait for the next tick
                        c = self._last_amount_sourcing_cores

                # but lets have a look at he mean (of the last 2 minutes) and current level of cache filling too
                if (
                    cache_level_mean > self.target_graphobjects_count_in_cache * 0.6
                    and cache_level_mean < self.target_graphobjects_count_in_cache * 1.2
                ):
                    # we are in the sweetspot of cache level. lets override the trend analyses result
                    # We can keep feeding the loaders but wont eat too much memory
                    # ❤️ lets keep beeing here happy forever ❤️
                    c = self._last_amount_sourcing_cores
                elif (
                    cache_level_mean > self.max_graphobjects_count_in_cache
                    or cache_level_current > self.max_graphobjects_count_in_cache
                ):
                    # we shot over the max level. no more new parsers until the situation is better.
                    c = 0

            else:
                c = self._last_amount_sourcing_cores
        max_c = math.floor(self.max_processes_count * 0.6)
        if c > max_c:
            c = max_c
        if c < 1:
            c = 1
        self._last_amount_sourcing_cores = c

        return c



"""Logging logger and handler via the neobulkmp cache backend
    """
import logging
import threading
from .cache_backend import CacheInterface

# Logger class on worker processes side. This serializes the log records and write them into the cache backend


class CacheLoggerHandler(logging.Handler):  # inherit from Handler class
    def __init__(self, cache: CacheInterface):
        logging.Handler.__init__(self)  # run parent __init__ class
        self.cache = cache

    def emit(self, record: logging.LogRecord):
        # Write log into caching backend
        self.cache.write_log(record)


def _handle_log_records(cache):
    for record in cache.fetch_logs():
        logger = logging.getLogger(record.name)
        logger.handle(record)


# Logger class on main process side. This grabs the serialized log records from the cache backend and logs them on the main process side


class LogRecordStreamHandler:
    """Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """

    def __init__(self, cache):
        self.cache = cache

    def handle(self):
        # _thread.start_new_thread(_test, ())
        # Todo: I dont know why the threading function does not work as local class method?
        # Did the same in https://git.connect.dzd-ev.de/dzdpythonmodules/ptan/-/blob/master/ptan/server.py#L159
        # for now we just outsource it to a standalone func '_handle_log_records'
        self.log_thread = threading.Thread(
            target=_handle_log_records, args=(self.cache,)
        )
        self.log_thread.daemon = True
        self.log_thread.start()

    def handleLogRecord(self, record: logging.LogRecord):
        logger = logging.getLogger(record.name)
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and bandwidth!
        logger.handle(record)



import multiprocessing
from .cache_backend import CacheInterface
from typing import Dict, List
from linetimer import CodeTimer
from .cache_logger import CacheLoggerHandler
import logging
import uuid
from enum import Enum


class Progress(str, Enum):
    # str mixin, to be able to compare enums to strings. cool trick! see https://stackoverflow.com/a/63028809/12438690
    QUEUED = "queued"
    DRAIN_ORDERED = "drain"
    DRAIN_READY = "drained"
    LOADING = "load"
    COMPLETE = "complete"


class WorkerBase(multiprocessing.Process):
    cache_backend: CacheInterface = None
    cache_backend_params: Dict = None

    def __init__(self):
        super(WorkerBase, self).__init__()
        self.tags: List[str] = []
        self.params: Dict = {}
        self.timer: CodeTimer = CodeTimer(silent=True, unit="s")
        self.cache: CacheInterface = None
        self.id: str = uuid.uuid4().hex
        self.progress: Progress = Progress.QUEUED

    def run(self):
        pass

    def get_logger(self) -> logging.Logger:
        if self.cache:
            logging.basicConfig(level=logging.INFO, handlers=[])
            log = logging.getLogger(self.name)
            log.addHandler(CacheLoggerHandler(self.cache))
            return log
        else:
            raise KeyError(
                "WorkerBase self.cache is None. Initate cache backed first before call 'get_logger'"
            )



import sys
import os

if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    MODULE_ROOT_DIR = os.path.join(SCRIPT_DIR, "..")
    sys.path.insert(0, os.path.normpath(MODULE_ROOT_DIR))
from neobulkmp import Manager, WorkerSourcing
import sys
import os
import logging
import graphio
import py2neo

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)-15s %(processName)-8s %(module)-8s %(levelname)-8s:  %(message)s",
    handlers=[logging.StreamHandler((sys.stdout))],
)

dbhost = os.getenv("CI_DB_HOST", "localhost")
g = py2neo.Graph(**{"host": dbhost})


def assert_result():
    r = g.run("MATCH (n:Actor) RETURN count(n) as cnt").data()[0]["cnt"]
    assert r == 12
    r = g.run("MATCH (n:Movie) RETURN count(n) as cnt").data()[0]["cnt"]
    assert r == 5
    r = g.run("MATCH (a:Actor)-[ai:ACTS_IN]->(m:Movie) RETURN count(ai) as cnt").data()[
        0
    ]["cnt"]
    assert r == 15


def clean_db():
    g.run("DROP INDEX Movie IF EXISTS")
    g.run("DROP INDEX Actor IF EXISTS")
    g.run("MATCH p=()-[:ACTS_IN]->() DELETE p")
    g.run("MATCH (a:Actor) DELETE a")
    g.run("MATCH (a:Movie) DELETE a")


def run_smoketest():

    movies = []
    movies.append(
        {
            "title": "Matrix",
            "Actors": ["Keanu Reeves", "Carrie-Anne Moss", "Laurence Fishburne"],
        }
    )
    movies.append(
        {
            "title": "Bill & Ted's Excellent Adventure",
            "Actors": ["Keanu Reeves", "Alex Winter", "George Carlin"],
        }
    )

    movies.append(
        {
            "title": "Ghost in the Shell",
            "Actors": ["Atsuko Tanaka", "Akio Otsuka", "Iemasa Kayumi"],
        }
    )

    movies.append(
        {"title": "Dogma", "Actors": ["Ben Affleck", "Jason Lee", "George Carlin"]}
    )
    movies.append(
        {
            "title": "Almost Famous",
            "Actors": ["Patrick Fugit", "Jason Lee", "Frances McDormand"],
        }
    )

    def parsing_func(worker: WorkerSourcing, title, Actors):
        # Create nodeset with movie
        log = worker.get_logger()
        movieset = graphio.NodeSet(["Movie"], merge_keys=["title"])
        movieset.add_node({"title": title})
        # Send Nodeset with movie to neobulkmp manager
        yield movieset
        # Create Actor Nodeset and connect actors to movie
        actorset = graphio.NodeSet(["Actor"], merge_keys=["name"])
        acts_in_movie = graphio.RelationshipSet(
            "ACTS_IN", ["Actor"], ["Movie"], ["name"], ["title"]
        )
        for actor in Actors:
            actorset.add_node({"name": actor})
            acts_in_movie.add_relationship({"name": actor}, {"title": title}, {})

        # Send actorset to neobulkmp manager
        yield actorset
        # Send actor to movie relation to neobulkmp manager
        yield acts_in_movie

    # Mount parsing function on parsing workers
    WorkerSourcing.sourcing_func = parsing_func

    # Init manager, provide parsing worker and worker parameters
    cachehost = os.getenv("CI_CACHE_HOST", "localhost")
    print("CACHE:HOST:", cachehost)
    manager = Manager(
        worker_sourcing_class=WorkerSourcing,
        worker_parameters=movies,
        cache_backend_params={"host": cachehost},
    )
    manager.strategy.max_graphobjects_count_in_cache = 2
    manager.manage_store_every_n_sec = 0.1
    manager.status_log_every_n_sec = 0.1
    manager.update_statistics_per_sec = 0.1

    # manager.max_concurrently_worker_count = 1
    manager.create_indexes = True
    dbhost = os.getenv("CI_DB_HOST", "localhost")
    # manager.create_unique_constraints = True  # not yet implemented
    manager.merge(graph_params={"host": dbhost})


# clean_db()
run_smoketest()
assert_result()



import logging
from typing import List, Dict, Type, Union, Tuple, Set, TYPE_CHECKING
from ..worker_base import WorkerBase, Progress
from multiprocessing.context import ProcessError

if TYPE_CHECKING:
    from .manager import Manager

log = logging.getLogger(__name__)


class ManagerWorkersBase:
    parent: "Manager" = None
    worker_class: Type[WorkerBase] = None

    def __init__(self, manager: "Manager", worker_class: Type[WorkerBase]):
        self.parent = manager
        self.workers: List[WorkerBase] = []
        self.finished_workers: List[WorkerBase] = []
        self.just_finished_workers: List[WorkerBase] = []
        self.failed_workers: List[WorkerBase] = []
        self.worker_class = worker_class

        # Setup caching backend on worker class

        self.worker_class.cache_backend = self.parent.cache_backend
        self.worker_class.cache_backend_params = self.parent.cache_backend_params

    def is_done(self):
        raise NotImplementedError

    def _get_process_status(self, worker: WorkerBase):
        if worker.is_alive():
            return "running"
        else:
            if worker.exitcode is None:
                return "initial"
            elif worker.exitcode == 0 and worker in self.workers:
                return "exited"
            elif worker.exitcode != 0 and worker in self.workers:
                return "failed"
            elif worker in self.finished_workers:
                return "closed"
        raise ValueError(
            "Could not determine process status. Something went wrong",
            worker.exitcode,
            worker,
        )

    def _init_workers(
        self, parameters: List[Dict], names: List[str] = None, tags: List = None
    ):
        if not names:
            names = [hash(str(p)) for p in parameters]
        for name, params in zip(names, parameters):
            name = f"{self.worker_class.__name__}-{';'.join(tags or [])}-{name}"
            w: WorkerBase = self.worker_class(name=name, **params)
            log.debug(f"INIT WORKER {w.name}")
            w.tags = tags
            w.params = params
            self.workers.append(w)

    def _finish_workers(self, tag: str = None) -> List[WorkerBase]:
        just_finished_workers = []
        for fin_worker in self._get_workers(status="exited", tag=tag):
            fin_worker.join()
            fin_worker.timer.__exit__(None, None, None)
            self.parent._release_loading_block_label(fin_worker.id)
            fin_worker.progress = Progress.COMPLETE
            self.finished_workers.append(fin_worker)
            just_finished_workers.append(fin_worker)
            log.debug(f"Exit worker '{fin_worker.name}'")
            # Free up resources
            del fin_worker

        # Exit failed worker
        for fail_worker in self._get_workers(status="failed", tag=tag):
            fail_worker.join()
            fail_worker.timer.__exit__(None, None, None)
            # release any nodeset blocks
            self.parent.blocked_labels.pop(fail_worker.id, None)
            self.failed_workers.append(fail_worker)
            just_finished_workers.append(fail_worker)
            self.parent._release_loading_block_label(fail_worker.id)
            fail_worker.progress = Progress.COMPLETE
            log.error(
                f"Exit failed worker '{fail_worker.name}' \n other workers currently running are {self._get_workers('running')}"
            )
            raise ProcessError(f"'{fail_worker.name}' failed")
        self.workers = [w for w in self.workers if w not in just_finished_workers]
        self.just_finished_workers += just_finished_workers
        return just_finished_workers

    def _get_workers(
        self,
        status: Union[str, List, Tuple, Set] = None,
        progress: Union[str, List, Tuple, Set] = None,
        tag=None,
    ) -> List[WorkerBase]:
        result = []
        if (status and (status == "closed" or "closed" in status)) or (
            progress
            and (progress == Progress.COMPLETE or Progress.COMPLETE in progress)
        ):
            result.extend(
                [
                    w
                    for w in self.finished_workers + self.failed_workers
                    if tag is None or tag in w.tags
                ]
            )
        result.extend(
            [
                w
                for w in self.workers
                if (tag is None or tag in w.tags)
                and (
                    status is None
                    or (
                        isinstance(status, str)
                        and self._get_process_status(w) == status
                    )
                    or (
                        isinstance(status, (list, set, tuple))
                        and self._get_process_status(w) in status
                    )
                )
                and (
                    progress is None
                    or (isinstance(progress, str) and w.progress == progress)
                    or (
                        isinstance(progress, (list, set, tuple))
                        and w.progress in progress
                    )
                )
            ]
        )
        return result



import graphio
from typing import Callable, Any, Union, Generator, Dict
from .worker_base import WorkerBase
from logging import Logger
from .cache_backend import CacheInterface
import time


class WorkerSourcing(WorkerBase):
    sourcing_func: Callable[
        [Any], Generator[Union[graphio.NodeSet, graphio.RelationshipSet], None, None]
    ] = None

    def __init__(self, name, **kwargs):
        super().__init__()
        self.name = name
        self.worker_parameter = kwargs

    def store_graphSet(self, graphset: Union[graphio.NodeSet, graphio.RelationshipSet]):
        data_stored = False
        while not data_stored:
            if self.cache.store_is_available():
                if isinstance(graphset, graphio.NodeSet):
                    self.log.debug(f"Recieved graphio.NodeSet '{graphset.labels}'")
                    self.cache.store_NodeSet(graphset)
                elif isinstance(graphset, graphio.RelationshipSet):
                    self.log.debug(
                        f"Recieved graphio.RelationshipSet '{graphset.rel_type}'"
                    )
                    self.cache.store_RelSet(graphset)
                else:
                    raise ValueError(
                        f"Worker '{self.name}' yielded wrong type of data. Expected 'graphio.NodeSet' or 'graphio.RelationshipSet' got '{type(graphset)}' "
                    )
                data_stored = True
            else:
                # wait 10 second for store to open again and try again
                time.sleep(1)

    def run(self):
        self.cache: CacheInterface = self.cache_backend(self.cache_backend_params)

        log: Logger = self.get_logger()
        self.log = log
        log.debug(f"START {self.name}")
        if not isinstance(self.worker_parameter, dict):
            raise ValueError("worker_parameter not a dict:", self.worker_parameter)
        for set in self.sourcing_func(**self.worker_parameter):
            self.store_graphSet(set)



# Demo Script for neobulkmp
# This Script will insert some movies with a sample of actors with multiple processes
# Requirements:
# * neobulkmp installed `pip install neobulkmp`
# * A local neo4j DB as database `docker run --rm --network host -e NEO4J_AUTH=none neo4j`
# * A local redis DB as a loading cache `docker run --rm --network=host redis`

from neobulkmp import Manager, WorkerSourcing
import sys
import os
import logging
import graphio
import dict2graph

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)-15s %(processName)-8s %(module)-8s %(levelname)-8s:  %(message)s",
    handlers=[logging.StreamHandler((sys.stdout))],
)


data = {
    "MedlineCitation": {
        "Status": "PubMed-not-MEDLINE",
        "Owner": "NLM",
        "PMID": {"Version": "1", "text": "20764441"},
        "DateCompleted": {"Year": "2011", "Month": "03", "Day": "29"},
        "DateRevised": {"Year": "2011", "Month": "03", "Day": "29"},
        "Article": {
            "PubModel": "Print",
            "Journal": {
                "ISSN": {"IssnType": "Print", "text": "0007-1447"},
                "JournalIssue": {
                    "CitedMedium": "Print",
                    "Volume": "1",
                    "Issue": "2522",
                    "PubDate": {"Year": "1909", "Month": "May", "Day": "01"},
                },
                "Title": "British medical journal",
                "ISOAbbreviation": "Br Med J",
            },
            "ArticleTitle": "BILATERAL NEPHRO-LITHOTOMY, IN WHICH THE KIDNEY WAS KEPT OUTSIDE THE WOUND FOR SEVEN DAYS BEFORE RETURNING IT TO THE LOIN.",
            "Pagination": {"MedlinePgn": "1059-60"},
            "AuthorList": {
                "CompleteYN": "Y",
                "Author": [
                    {
                        "ValidYN": "Y",
                        "LastName": "Clay",
                        "ForeName": "J",
                        "Initials": "J",
                    }
                ],
            },
            "Language": ["eng"],
            "PublicationTypeList": {
                "PublicationType": [{"UI": "D016428", "text": "Journal Article"}]
            },
        },
        "MeshHeadingList": {
            "MeshHeading": [
                {
                    "DescriptorName": {
                        "UI": "D000311",
                        "MajorTopicYN": "N",
                        "text": "Adrenal Glands",
                    },
                    "QualifierName": [
                        {"UI": "Q000601", "MajorTopicYN": "Y", "text": "surgery"}
                    ],
                },
                {
                    "DescriptorName": {
                        "UI": "D013507",
                        "MajorTopicYN": "Y",
                        "text": "Endocrine Surgical Procedures",
                    }
                },
                {
                    "DescriptorName": {
                        "UI": "D006801",
                        "MajorTopicYN": "N",
                        "text": "Humans",
                    }
                },
                {
                    "DescriptorName": {
                        "UI": "D008216",
                        "MajorTopicYN": "Y",
                        "text": "Lymphocytic Choriomeningitis",
                    }
                },
                {
                    "DescriptorName": {
                        "UI": "D008581",
                        "MajorTopicYN": "Y",
                        "text": "Meningitis",
                    }
                },
            ]
        },
        "MedlineJournalInfo": {
            "Country": "England",
            "MedlineTA": "Br Med J",
            "NlmUniqueID": "0372673",
            "ISSNLinking": "0007-1447",
        },
    },
    "PubmedData": {
        "History": {
            "PubMedPubDate": [
                {
                    "PubStatus": "entrez",
                    "Year": "2010",
                    "Month": "8",
                    "Day": "27",
                    "Hour": "6",
                    "Minute": "0",
                },
                {
                    "PubStatus": "pubmed",
                    "Year": "1909",
                    "Month": "5",
                    "Day": "1",
                    "Hour": "0",
                    "Minute": "0",
                },
                {
                    "PubStatus": "medline",
                    "Year": "1909",
                    "Month": "5",
                    "Day": "1",
                    "Hour": "0",
                    "Minute": "1",
                },
            ]
        },
        "PublicationStatus": "ppublish",
        "ArticleIdList": {
            "ArticleId": [
                {"IdType": "pubmed", "text": "20764441"},
                {"IdType": "pmc", "text": "PMC2318724"},
            ]
        },
    },
}
"""
data = {
    "MedlineCitation": {
        "Status": "PubMed-not-MEDLINE",
        "Owner": "NLM",
        "PMID": {"Version": "1", "text": "20764441"},
        "DateCompleted": {"Year": "2011", "Month": "03", "Day": "29"},
        "DateRevised": {"Year": "2011", "Month": "03", "Day": "29"},
        "Article": {
            "PubModel": "Print",
            "Journal": {
                "ISSN": {"IssnType": "Print", "text": "0007-1447"},
                "JournalIssue": {
                    "CitedMedium": "Print",
                    "Volume": "1",
                    "Issue": "2522",
                    "PubDate": {"Year": "1909", "Month": "May", "Day": "01"},
                },
                "Title": "British medical journal",
                "ISOAbbreviation": "Br Med J",
            },
        },
        "AuthorList": {
            "CompleteYN": "Y",
            "Author": [
                {
                    "ValidYN": "Y",
                    "LastName": "Clay",
                    "ForeName": "J",
                    "Initials": "J",
                }
            ],
        },
    }
}
"""


class MyCustomWorker(WorkerSourcing):
    def sourcing_func(worker, srcdata):
        d2g = dict2graph.Dict2graph()
        d2g.parse(srcdata, "Movie", instant_save=False)
        d2g.save()
        for nodeSet in d2g.nodeSets.values():
            yield nodeSet
            # worker.cache.store_NodeSet(nodeSet)

        for relSet in d2g.relationshipSets.values():
            yield relSet
            # worker.cache.store_RelSet(relSet)
        d2g.clear()


# Init manager, provide parsing worker and worker parameters
manager = Manager(
    worker_sourcing_class=MyCustomWorker, worker_parameters=[{"srcdata": data}]
)
# manager.max_concurrently_worker_count = 1
manager.create_indexes = True
# manager.create_unique_constraints = True  # not yet implemented
manager.merge(graph_params={"host": "localhost"})
